# -*- coding: UTF-8 -*-
"""
05/04/2018: for infea_clean/messy test
			This version is mainly for feasibility_primal test (mode 3)
			We aim to test for infeasibility: either strongly infeasible (Case 2 in our paper) or weakly infeasible (Case 4 in our paper)
			If the optimal value from mode 3 >= eps (right now set it to be 1.e-20):
				- Case 2: Strongly infeasible
			If the optimal value from mode 3 <= -eps:
				- Case 1: Strictly feasible
			If the optimal value from mode 3 is between (-eps, eps), we numerically treat it as 0, and then we have two sub-cases:
				(1). optimal value is attained (this means X must be numerically real, for the dual variables, they are theoretically real anyway!)
					- Case 3: suspected of only feasible
				(2). optimal value is NOT attained for X, i.e., X is complex, then
					- Case 4: suspected of weakly infeasible
			Remark: the choice of eps will affect the output the status of the feasibility test!
			Note that if test_linear_depend fails, then our code cannot handle this!
			We assume the test cases from http://gaborpataki.web.unc.edu/infeasible-and-weakly-infeasible-sdps/
				will satisfies test_linear_depend and m <= n*(n+1)/2, otherwise further investigation is needed!
			Also, in our mode 3, we don't need the origional info. for C matrix since our homotopy formulation does not need it!
			In Liu and Pataki's exact duals paper, they set C = I_10 (identity with 10*10 matrix). Anyway we still read C matrix as I_10, but it will be modified to zero matrix in code.
"""
import collections
import os
import numpy as np
import scipy as sp
import sys

import bertini
import sdp_in

# for convenience
def print_data(C, A, b):
	print('')
	print(type(C))
	print(C)
	for i in range(len(A)):
		print(type(A[i]))
		print(A[i])
		print(type(b))
		print(b)

def test_linear_depend(A, *args):
	"""
	test linear dependence between A_i and possible *args
	use: numpy.linalg.matrix_rank(M, tol=None)
	"""
	m = len(A)
	n = A[0].shape[0] # square
	M = np.zeros((m,int(n*(n+1)/2))) # row-major data for efficiency?
	for i in range(m):
		M[i,:] = A[i][np.triu_indices(n)]
	for item in args: # args is a tuple
		M = np.vstack([M, item[np.triu_indices(n)]])
		m += 1
	rank = np.linalg.matrix_rank(M)
	r = min(m, int(n*(n+1)/2))

	return rank < r # if this True, then linear dependent

def test_feasibility_equality(A,b):
	"""
	test feasibility of A_i \cdot X = b_i, i = 1,...,m
	Since A_i are already linearly independent, it's not necessary to solve them directly!
	"""
	m = len(A)
	n = A[0].shape[0] # square
	N = int(n*(n+1)/2)
	feasible = False
	if m<=N:
		print("m <= n(n+1)/2, underdetermined or square system for equality constraints")
		print("Since A_i are already linearly independent")
		print("A_i \cdot X = b_i, i = 1,...,m is always feasible!")
		print("Continue...")
	elif m>N:
		print("Warning: m > n(n+1)/2, overdetermined system for equality constraints")
		print("Warning: waiting to be tested!")
		M = np.zeros((m,N))
		for i in range(m):
			idx = 0
			for j in range(n):
				for k in range(j,n):
					if j==k:
						M[i,idx] = A[i][j,k]
					else:
						M[i,idx] = 2*A[i][j,k]
					idx += 1
			if idx != N-1:
				print("idx error!")
				exit()
		x_lstsq = np.linalg.lstsq(A,b)[0]
		residual = np.linalg.norm(np.dot(M,x_lstsq) - b)
		print("residual = {0}".format(residual))
		tol = 1.0e-8
		if abs(residual) > tol:
			print("residual too big, exit")
			exit()

def find_start_points_optimum(C, A, b, mode):
	# reshape feasible set
	print('---Inside find_start_points_optimum---')
	m = len(b)
	n = C.shape[0] # square
	ys = [np.random.choice((-1,1))*np.random.random() for i in range(m)] # len(ys) == m

	if mode == 2:
		ys[m-1] = 1 # the last one is modified to be 1, M = 2
	Ss = C - sum([ys[i]*A[i] for i in range(m)], np.zeros((n,n))) # why need np.zeros((n,n))?

	# choose \tilde{y}:=ysmp1 to make S* strictly positive definite
	smallest_eig = min(np.linalg.eigvals(Ss))
	#print('smallest_eig = ', smallest_eig)
	if smallest_eig <= 0:
		ysmp1 = smallest_eig - 1
		Amp1 = np.eye(n)
	else:
		ysmp1 = 0
		Amp1 = np.zeros((n,n))
	if mode == 2: # feasibility_dual mode
		if (ys[m-1]+ysmp1) > 0: # ys[m-1] is origional ys_{m+1}
			print('ys[m-1]+ysmp1 = {0} > 0 and hence SDP-D is strictly feasible'.format(ys[m-1]+ysmp1))
			print('Early exit in ' + find_start_points_optimum.__name__)
			print(ys)
			#print('Still continue, ok! If only do the SDP-D feasibility test, we can exit here!')
			exit()
	
	Ss -= ysmp1*Amp1 # make it to be "-"
	#print('final smallest eig of Ss = ', min(np.linalg.eigvals(Ss)))
	#print(np.linalg.eigvals(Ss))
	Xs = np.linalg.inv(Ss) # in mode 3, Xs is  actually \tilde{Xs}

	bs = [(A[i]*Xs).sum() for i in range(m)]
	if mode == 2: # modify bs[m-1]
		# beta_s = 1 since it is (M - ys[m-1])^{-1}, M = 2 and ys[m-1] = 1
		bs[m-1] = np.trace(Xs) + 1 # it seems that we do not need this info.
	if mode == 3: # len(ys) = m+1 now since ys[m] = \lambda
		# enlarge ys for mode 3 (feasibility_primal)
		#ys.append(np.random.choice((-1,1))*np.random.random())
		# Set M = 2, choose gamma_s=1>0, then lambda_s = -1
		ys.append(-1)
		# now len(ys) = m+1, the last variable represents lambda
		bs = [bs[i] - ys[m]*np.trace(A[i]) for i in range(m)]
	trace_Ss = np.trace(Ss)

	# ysmp1 is \tilde{y}
	if mode == 3:
		return Xs,ys,bs,ysmp1,trace_Ss # for mode 3, len(ys) = m+1 instead of m
	else:
		return Xs,ys,bs,ysmp1

def compute_optimum(C, A, b, mode, foldername, example_tag):
	# compute start points
	# C, A are np.2darray (full dense matrix), b is a list of float
	# Xs: np.ndarray, ys, bs: list, ysmp1 = scaler (nonpositive!)
	# print_data(C,A,b)
	print("---Inside compute_optimum---")
	if mode == 3:
		Xs,ys,bs,ysmp1,trace_Ss = find_start_points_optimum(C, A, b, mode)
	else: # mode == 1 or 2
		Xs,ys,bs,ysmp1 = find_start_points_optimum(C, A, b, mode)
	m = len(b) # use this, b/c in mode 3, len(y)=m+1 (adding \lambda to the last position of y)
	n = C.shape[0] # square

	# register constants with bertini
	constants = collections.OrderedDict()
	for i in range(m): # only input upper tri. of Ai
		for j in range(n):
			for k in range(j,n):
				constants["A{0}_{1}_{2}".format(i,j,k)] = A[i][j,k]
	for j in range(n): # only input upper tri. of C
		for k in range(j,n):
			constants["C{0}_{1}".format(j,k)] = C[j,k]
	for i in range(m): # bs stands for b*
		constants["b{0}".format(i)] = b[i]
		constants["bs{0}".format(i)] = bs[i]

	#constants["y{0}".format(m)] = ymp1
	constants["ysmp1"] = ysmp1 # use ysmp1:=\tilde{y} bc y_m is reserved for \lambda in mode 3

	# two sets of variables: dim(y_var)=m, dim(X_var)=(n+1)*n/2
	yvariables = []
	Xvariables = []

	# form base homotopy
	# subfunction: S = C - \sum_i A_i * y_i - ysmp1 * \mu * Id, upper tri.
	# append X variables, upper tri.
	# here parameters={'mu':'t'}
	subfunctions = collections.OrderedDict()
	for j in range(n):
		for k in range(j,n):
			Xvariables.append("X{0}_{1}".format(j,k))
			eqterms = ["C{0}_{1}".format(j,k)]
			for i in range(m):
				eqterms.append("A{0}_{1}_{2}*y{0}".format(i,j,k))
			g = " - ".join(eqterms)
			if j == k:
				g += " - ysmp1*mu" # make it to be "- ysmp1 * mu"
			subfunctions["S{0}_{1}".format(j,k)] = g

	# two set of eqns
	functions = collections.OrderedDict()

	# first set (m eqns): A_i \cdot X - b_i * (1-mu) - bs_i * mu
	# append y variables
	if mode == 1 or mode == 3:
		mm = m
	if mode == 2:
		mm = m-1 # cut off the last one
	for i in range(mm):
		eqterms = []
		yvariables.append("y{0}".format(i))
		for j in range(n):
			eqterms.append("A{0}_{1}_{1}*X{1}_{1}".format(i,j))
			for k in range(j+1,n):
				eqterms.append("2*A{0}_{1}_{2}*X{1}_{2}".format(i,j,k))
		functions["f{0}".format(i)] = " + ".join(eqterms) + " - b{0}*(1 - mu) - bs{0}*mu".format(i)
	if mode == 2: # add back the m position of the equality constraint
		yvariables.append("y{0}".format(m-1))
		eqterms = []
		for i in range(n):
			eqterms.append("X{0}_{0}".format(i))
		# M = 2
		functions["f{0}".format(m-1)] = "( 1 + mu * {0} - ".format(np.trace(Xs)) + "-".join(eqterms) + ") * (2-y{0}) - mu".format(m-1)

	# modify first set of eqns. for mode 3
	if mode == 3:
		yvariables.append("y{0}".format(m)) # add variable lambda to yvariables y[m]
		for i in range(m):
			functions["f{0}".format(i)] += " - y{0} * {1}".format(m,np.trace(A[i]))

	# for mode 3, add one more eqn to f[m]: XXXXXX S \cdot I - 1*(1-\mu) - \mu * trace(Ss) = 0
	# New: (1 + \mu trace(Ss) - trace(S)) * (\lambda + M) - \mu = 0
	if mode == 3:
		eqterms = [] # waster a lot of old eqterms??? modify later for efficiency
		for i in range(n):
			eqterms.append("S{0}_{0}".format(i))
		#["f{0}".format(m)] = " + ".join(eqterms) + " - (1 - mu) - mu * {0}".format(trace_Ss) 
		# " - 1 - mu * {0}".format(trace_Ss-1.0)
		functions["f{0}".format(m)] = "(" + "1 + mu * {0}".format(trace_Ss) + " - " + " - ".join(eqterms) + ")" + "* (y{0} + 2) - mu".format(m)

	# second set ((n+1)*n/2 eqns): S.dot(X) - \mu * Id
	# only upper tri. part of S and X
	# symmetric version!
	for i in range(n):
		eqterms = []
		for k in range(n):
			if i <= k:
				eqterms.append("2*S{0}_{1}*X{0}_{1}".format(i, k))
			else:
				eqterms.append("2*S{0}_{1}*X{0}_{1}".format(k, i))
		h = " + ".join(eqterms) + " - 2*mu"
		functions["h{0}_{0}".format(i)] = h

		for j in range(i+1, n):
			eqterms = []
			for k in range(n):
				if i <= k and k <= j:
					eqterms.append("S{0}_{1}*X{1}_{2} + X{0}_{1}*S{1}_{2}".format(i,k,j))
				elif i <= k and k > j:
					eqterms.append("S{0}_{1}*X{2}_{1} + X{0}_{1}*S{2}_{1}".format(i,k,j))
				elif i > k and k <= j:
					eqterms.append("S{1}_{0}*X{1}_{2} + X{1}_{0}*S{1}_{2}".format(i,k,j))
				else:
					eqterms.append("S{1}_{0}*X{2}_{1} + X{1}_{0}*S{2}_{1}".format(i,k,j))
				h = " + ".join(eqterms)
			functions["h{0}_{1}".format(i,j)] = h

	#variable_group = Xvariables + yvariables
	## make variable_group to be list of list
	if mode == 3:
		variable_group = [Xvariables, yvariables[:len(yvariables)-1], [yvariables[len(yvariables)-1]]]
	elif mode == 2:
		variable_group = [Xvariables, yvariables[:len(yvariables)-1], [yvariables[len(yvariables)-1]]] # the last is y_{m+1}
	else: # mode == 1:
		variable_group = [Xvariables, yvariables]

	# write out the input files
	dirname = "brun/"
	foldername = foldername
	dirname = os.path.join(dirname, foldername)
	dirname = os.path.join(dirname, example_tag)
	try:
		os.mkdir(dirname)
	except OSError as e:
		if e.errno == 17: # file exists
			pass
		else:
			raise(e)

	Xpoints = []
	#Spoints = [] ## Questions: do we need Spoints to start?
	for i in range(n):
		for j in range(i, n):
			Xpoints.append(Xs[i,j])
			#Spoints.append(Ss[i,j])

	#startpoints = [bertini.Point(Xpoints + Spoints + ys)]
	startpoints = [bertini.Point(Xpoints + ys)]
	bertini.write_bertini_input_file(dirname, variable_group, constants, subfunctions, functions, parameters={'mu':'t^8'}, pathvariables=['t'], options={"UserHomotopy":2, "SecurityLevel":1, "TrackTolBeforeEG":1e-9, "TrackTolDuringEG":1e-9, "ODEPredictor":2, "SharpenDigits":50})
	bertini.write_bertini_start_file(dirname, startpoints)
	bertini.run_bertini(dirname, "optimum") # optimum mode

	try:
		optimum = bertini.read_solutions_file(dirname, "nonsingular_solutions")[0] # only one of these
	except ZeroDivisionError:
		optimum = bertini.read_solutions_file(dirname, "singular_solutions")[0] # only one of these

	Xcoords = optimum.coordinates[:int(n*(n+1)/2)]
	#Scoords = optimum.coordinates[int(n*(n+1)/2):n*(n+1)]
	ycoords = optimum.coordinates[int(n*(n+1)/2):]
	
	tol = 1.0e-10 # change here if necessary
	Xcoords_imag = [z.imag for z in Xcoords]
	normX_imag = np.linalg.norm(Xcoords_imag, np.inf)
	ycoords_imag = [z.imag for z in ycoords]
	normy_imag = np.linalg.norm(ycoords_imag, np.inf)
	Ximag_flag = False
	if normX_imag > tol:
		print("\nWarning: non-real solution for X!")
		Ximag_flag = True
	yimag_flag = False
	if normy_imag > tol:
		print("\nWarning: non-real solution for y!") # if y is real, then S is of course real!
		yimag_flag = True
	if max(normX_imag,normy_imag) > tol:
		print("Output info. in postprocessing stage maybe incorrect!")
		print("Go to check solution files for more info!")

	X = np.matrix(np.zeros((n,n)))
	S = np.matrix(np.zeros((n,n)))
	k = 0
	y = np.array([z.real for z in ycoords])
	for i in range(n):
		for j in range(i,n):
			X[i,j] = X[j,i] = Xcoords[k].real
			S[i,j] = S[j,i] = C[i,j] - sum([y[l]*A[l][i,j] for l in range(m)])
			k += 1
	
	if mode == 3:
		return X, y, S, Ximag_flag, yimag_flag # return np.matrix, np.array
		# in mode 3, y[m] is lambda!
	else:
		return X, y, S

def compute_feasibility_dual(C, A, b, mode):
	# max y_{m+1} s.t. C - \sum Ai*yi - I*y_{m+1} >= 0
	# => b~ = [0;...;0;1] \in \R^{m+1}
	## enlarge b and A, A[m+1] = I

	print('---Inside compute_feasibility_dual---')
	n = C.shape[0] # square
	# check whether I is a linear combination of A_i's
	# if so, we come to a conclusion: SDP-D is strictly feasible (can be proved straightforwardly!)
	# no need to continue computing in this case!
	# If C > 0, then it is strict feasible immediately (set y_i = 0 for all i)
	linear_depend_info = test_linear_depend(A, np.eye(n))
	if linear_depend_info:
		print('Identity matrix is a linear combination of A_i (i=1,...,m)')
		print('Hence SDP-D is strictly feasible!')
		print('Early exit in ' + compute_feasibility_dual.__name__)
		#print('----')
		exit()
	else:
		print('We can safely append A with I and modify b, continue...')
	print('---Continue---')

	# modify b, add one more dimension, but this change only affect inside comp_feasibiility_dual, why???
	b = [0]*len(b) + [1] # make it to be "+"
	# modify A, add one more dimension
	A.append(np.eye(n)) # this modification will affect outside this function if A (not A[:]) is passed in
	X,y,S = compute_optimum(C,A,b,mode)

	return X,y,S # len(y) == m+1

def compute_feasibility_primal(C, A, b, mode, foldername, example_tag):
	"""
	min \lambda { X + \lambda I >= 0, A_i \cdot X = b_i, i=1,...,m }
	add one more primal variable \lambda, this new variable will append to the last of yvariables
	this will bring one more constraint eqn. into the KKT system: S \cdot I - 1 = 0
	Set C = 0
	Remark: note that here we use X to represent X + \lambda I, hence we need to modify the first set of eqns.
	"""
	print('---Inside compute_feasibility_primal---')
	n = C.shape[0] # square
	C = np.zeros((n,n)) # or change in the homotopy section to remove C-info.
	# !!! Set C = 0 to reuse the code, and for correctness, C MUST be 0 in mode 3!

	X, y, S, Ximag_flag, yimag_flag = compute_optimum(C,A,b,mode,foldername,example_tag)

	return X,y,S,Ximag_flag,yimag_flag # y[m] is lambda in mode 3!

def postprocess(X,y,S,C,A,b):
	"""
	Postprocessing stage:
	 1. check whether the solu. is feasible
	 2. check duality gap
	 3. provide info. for feasibility test
	 4. fix: C,A[i]: ndarray, b: list
	"""
	tol = 1.0e-14 ## change here if necessary
	m = len(b)
	n = C.shape[0] # square
	# convert list b to np.array, C, A[i] np.ndarry -> np.matrix
	C = np.matrix(C)
	for i in range(m):
		A[i] = np.matrix(A[i])
	b = np.asarray(b)
	
	# X feasible?
	smallest_eig_X = min(np.linalg.eigvals(X))
	print('smallest_eig_X = {0}'.format(smallest_eig_X))
	if smallest_eig_X >= -tol:
		print('X is spd and hence primal feasible!')
	else:
		print('X is NOT spd!')

	# trace(A_i .* X) - b_i = 0?
	test_equality_AXb = max([abs((np.multiply(A[i],X)).sum() - b[i]) for i in range(m)])
	if test_equality_AXb <= tol:
		print('trace(Ai * X) - bi ~= 0 for all i')
	else:
		print('trace(Ai * X) - bi ~= 0 for some i')

	# S*X = 0?
	test_equality_SX = (S*X).max()
	if test_equality_SX <= tol:
		print('S * X = 0')
	else:
		print('S * X ~= 0')

	# S feasible?
	smallest_eig_S = min(np.linalg.eigvals(S))
	print('\nsmallest_eig_S = {0}'.format(smallest_eig_S))
	if smallest_eig_S >= -tol:
		print('S is spd and hence dual feasible!')
	else:
		print('S is NOT spd!')

	# duality gap: C.*X - np.dot(b,y)
	# C*X: matrix multiplication, np.multiply(C,X): elem-wise mult.
	primal_obj_value = (np.multiply(C,X)).sum()
	print('\nprimal objective value = {0}'.format(primal_obj_value))
	dual_obj_value = np.dot(b, y)
	print('dual objective value = {0}'.format(dual_obj_value))
	print('duality gap = {0}'.format(primal_obj_value - dual_obj_value))

# call: python 800test.py > output.txt
if __name__ == '__main__':
	## set up parameters here
	mode_dict = { '1':'optimum_solve', '2':'feasibility_test_dual', '3':'feasibility_test_primal', '4':'other test' }
	print(sys.argv)
	#example_tag = sys.argv[1]
	mode = 3
	if mode == 3:
		#foldername = 'infeas_messy_20_10' # modify here, format: weak_clean_m_n
		foldername = sys.argv[1]
		print('---------------mode {0}: {1}-------------------'.format(mode, mode_dict[str(mode)]))
		print(foldername)

	cnt = 0
	faillist = []
	for k in range(0,100):
		kk = k + 1
		example_tag = str(kk)
		print('---------------Example {0}-------------------'.format(example_tag))
		# read data from file located in examples: C, A, b
		cwd = os.getcwd()
		example_dirname = os.path.join(cwd, 'examples_infeasible/'+foldername)
		C,A,b = sdp_in.read_in_SDP_infeas_test(example_dirname, foldername, example_tag)
		# task 0: first check linear dependence of A_i
		# this step can be skipped if A_i are already linearly indepenent
		# but there is no harm doing it for small test problem
		linear_depend_info = test_linear_depend(A)
		if linear_depend_info:
			print('Please remove redundacy between A_i (i=1,...,m) before running this code!')
			print('Early exit in ' + __name__)
			print('Others')
			faillist.append(kk)
			continue
			#exit()

		# test feasibility of A_i \cdot X = b_i
		test_feasibility_equality(A,b)

		# feasibility_primal
		X,y,S,Ximag_flag,yimag_flag = compute_feasibility_primal(C, A[:], b, mode, foldername, example_tag)
		print('\n-Output: feasibility Test for SDP_P-')
		if yimag_flag == True:
			print("y is complex, strange, STOP and check it!")
			print('Others')
			print('\n')
			#exit()
			faillist.append(kk)
			continue
		n = C.shape[0]
		m = len(b) # len(y) == m+1
		lambd = y[m]
		gamma = 1 - np.trace(S)
		print("gamma = {0}---should be nonnegative!".format(gamma))
		optimal_dual_value = np.dot(b, y[:m]) - gamma*2 # M = 2
		print('opt_optim_value = {0}, opt_dual_value = {1}'.format(lambd, optimal_dual_value))
		duality_gap = lambd - optimal_dual_value
		print("duality gap = {0}".format(duality_gap))
		eps_duality = 1.0e-9
		print('tolerance: eps_duality = {0}'.format(eps_duality))
		if duality_gap > eps_duality:
			print("Almost nonzero duality gap: strange, STOP and check it!")
			print('Others')
			faillist.append(kk)
			continue 
		eps = 1.0e-20
		print('tolerance: eps = {0}'.format(eps))
		if Ximag_flag == True and abs(lambd) < eps:
			print("optimal value from feasibility_primal test is almost 0 and X is complex, we could guess:")
			print("Case 4: (SDP-P) is suspected of weakly infeasible!")
			print("\n")
			cnt = cnt + 1
		elif Ximag_flag == False: # X is "real"
			#faillist.append(kk)
			if lambd <= -eps:
				print("Case 1: (SDP-P) is strictly feasible!\n")
				faillist.append(kk)
				#continue
			elif lambd >= eps:
				print("Case 2: (SDP-P) is strongly infeasible!\n")
				cnt = cnt + 1
			else: # ==0?, lambd is in (-eps, eps)
				#print("Two possible cases for Primal of SDP:\n feasible, but not strict feasible \n OR infeasible")
				print("It is safer to say: (SDP-P) with X + eps * I >= 0 is feasible, and (SDP-P) with X - eps * I >= 0 is infeasible!")
				print("Case 3: Since optimal value of feasibility_primal test is between (-eps, eps), ")
				print("we could guess: (SDP-P) is suspected of only feasible!\n")
				faillist.append(kk)
		else:
			faillist.append(kk)
			print('Others')
		print('\n')

	print('*********************')
	print('Success ration: {0}'.format(float(cnt)/float(100)))
	print(faillist)
	print('\n')