% Proof-of-concept implementation of the SDP solver in 
% J. Hauenstein, A. Liddell, S. McPherson, and Y. Zhang.
% Numerical algebraic geometry and semidefinite programming.
% Copyright 2021

function [OutcomeP,X,PO,OutcomeD,S,y,DO] = SDPsolver(A,b,C,primalExpT,dualExpT,primalDualExpT)
% INPUT: A = k x n^2 matrix (k = # of linear equations, X is n x n matrix)
%        b = k vector
%        C = n^2 vector
% OUTPUT: OutcomeP = 0 if primal is strictly feasible, 1 if primal is
%                    feasible but not strictly feasible, 2 if primal is infeasible
%         X = primal solution
%         PO = primal optimal value
%         OutcomeD = 0 if dual is strictly feasible, 1 if dual is
%                    feasible but not strictly feasible, 2 if dual is infeasible
%         S,y = dual solution
%         DO = dual optimal value

    % perform feasibility tests
    [OutcomeP,X] = PrimalFeasibility(A,b,C,primalExpT);
    [OutcomeD,S,y] = DualFeasibility(A,b,C,dualExpT);
    
    % proceed depending on feasibility
    if (OutcomeP == 0 && OutcomeD == 0) || (OutcomeP == 0 && OutcomeD == 1) || (OutcomeP == 1 && OutcomeD == 0)
        % both are feasible and at least one is strictly feasible
        [~,X,S,y,PO,DO] = HomPrimalDualSolver(A,b,C,primalDualExpT);
        if OutcomeP == 1 && OutcomeD == 0
            % dual optimal value is always attained
            PO = DO;  
        elseif OutcomeP == 0 && OutcomeD == 1
            % primal optimal value is always attained
            DO = PO;
        end;
    elseif OutcomeP == 2 && OutcomeD == 2
        % both are infeasible
        DO = -inf;
        PO = inf;
    elseif OutcomeP == 0 && OutcomeD == 2
        % primal is strictly feasible and dual is infeasible
        DO = -inf;
        PO = -inf;
    elseif OutcomeP == 2 && OutcomeD == 0
        % dual is strictly feasible and primal is infeasible
        DO = inf;
        PO = inf;
    elseif OutcomeP == 1 && OutcomeD == 1
        % both are feasible but not strictly feasible

        % primal facial reduction and solve
        [Aout,bout,Cout] = PrimalFacialReduction(X,A,b,C);
        [~,X,~,~,PO,~] = HomPrimalDualSolver(Aout,bout,Cout,primalDualExpT);
        
        % dual facial reduction and solve
        [Aout,bout,Cout,ObjectiveValueConstant] = DualFacialReduction(S,y,A,b,C);
        [~,~,S,y,~,DO] = HomPrimalDualSolver(Aout,bout,Cout,primalDualExpT);
        DO = DO + ObjectiveValueConstant;
    elseif OutcomeP == 1 && OutcomeD == 2
        % primal is feasible but not strictly feasible and dual is infeasible
        DO = -inf;
        % primal facial reduction and solve
        [Aout,bout,Cout] = PrimalFacialReduction(X,A,b,C);
        [~,X,~,~,PO,~] = HomPrimalDualSolver(Aout,bout,Cout,primalDualExpT);
    elseif OutcomeP == 2 && OutcomeD == 1
        % dual is feasible but not strictly feasible and primal is infeasible
        PO = inf;
        % dual facial reduction and solve
        [Aout,bout,Cout,ObjectiveValueConstant] = DualFacialReduction(S,y,A,b,C);
        [~,~,S,y,~,DO] = HomPrimalDualSolver(Aout,bout,Cout,primalDualExpT);
        DO = DO + ObjectiveValueConstant;
    end;
end
